from importlib import import_module
from types import ModuleType
from typing import List, Tuple

from injector import Module, multiprovider, provider, singleton

from isar.apis.api import API
from isar.apis.schedule.drive_to import DriveTo
from isar.apis.schedule.start_mission import StartMission
from isar.apis.schedule.stop_mission import StopMission
from isar.apis.security.authentication import Authenticator
from isar.config import config
from isar.config.keyvault.keyvault_service import Keyvault
from isar.mission_planner.echo_planner import EchoPlanner
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.mission_planner_interface import MissionPlannerInterface
from isar.models.communication.queues.queues import Queues
from isar.services.service_connections.mqtt.mqtt_client import (
    MqttClient,
    MqttClientInterface,
)
from isar.services.service_connections.request_handler import RequestHandler
from isar.services.service_connections.stid.stid_service import StidService
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine
from isar.storage.blob_storage import BlobStorage
from isar.storage.local_storage import LocalStorage
from isar.storage.slimm_storage import SlimmStorage
from isar.storage.storage_interface import StorageInterface
from robot_interface.robot_interface import RobotInterface


class APIModule(Module):
    @provider
    @singleton
    def provide_api(
        self,
        authenticator: Authenticator,
        start_mission: StartMission,
        stop_mission: StopMission,
        drive_to: DriveTo,
    ) -> API:
        return API(authenticator, start_mission, stop_mission, drive_to)

    @provider
    @singleton
    def provide_drive_to(self, scheduling_utilities: SchedulingUtilities) -> DriveTo:
        return DriveTo(scheduling_utilities)

    @provider
    @singleton
    def provide_start_mission(
        self,
        mission_planner: MissionPlannerInterface,
        scheduling_utilities: SchedulingUtilities,
    ) -> StartMission:
        return StartMission(mission_planner, scheduling_utilities)

    @provider
    @singleton
    def provide_stop_mission(self, queues: Queues) -> StopMission:
        return StopMission(queues)


class AuthenticationModule(Module):
    @provider
    @singleton
    def provide_authenticator(self) -> Authenticator:
        return Authenticator()


class RobotModule(Module):
    @provider
    @singleton
    def provide_robot_interface(self) -> RobotInterface:
        robot_package_name: str = config.get("DEFAULT", "robot_package")
        robot: ModuleType = import_module(robot_package_name)
        return robot.robotinterface.Robot()  # type: ignore


class QueuesModule(Module):
    @provider
    @singleton
    def provide_queues(self) -> Queues:
        return Queues()


class RequestHandlerModule(Module):
    @provider
    @singleton
    def provide_request_handler(self) -> RequestHandler:
        return RequestHandler()


class BlobStorageModule(Module):
    @multiprovider
    @singleton
    def provide_blob_storage(self, keyvault: Keyvault) -> List[StorageInterface]:
        return [BlobStorage(keyvault)]


class LocalStorageModule(Module):
    @multiprovider
    @singleton
    def provide_local_storage(self) -> List[StorageInterface]:
        return [LocalStorage()]


class SlimmStorageModule(Module):
    @multiprovider
    @singleton
    def provide_slimm_storage(
        self, request_handler: RequestHandler
    ) -> List[StorageInterface]:
        return [SlimmStorage(request_handler=request_handler)]


class LocalPlannerModule(Module):
    @provider
    @singleton
    def provide_local_planner(self) -> MissionPlannerInterface:
        return LocalPlanner()


class EchoPlannerModule(Module):
    @provider
    @singleton
    def provide_echo_planner(
        self,
        request_handler: RequestHandler,
        stid_service: StidService,
    ) -> MissionPlannerInterface:
        return EchoPlanner(
            request_handler=request_handler,
            stid_service=stid_service,
        )


class StateMachineModule(Module):
    @provider
    @singleton
    def provide_state_machine(
        self,
        queues: Queues,
        robot: RobotInterface,
        mqtt_client: MqttClientInterface,
    ) -> StateMachine:
        return StateMachine(queues=queues, robot=robot, mqtt_client=mqtt_client)


class UtilitiesModule(Module):
    @provider
    @singleton
    def provide_scheduling_utilities(self, queues: Queues) -> SchedulingUtilities:
        return SchedulingUtilities(queues)


class ServiceModule(Module):
    @provider
    @singleton
    def provide_keyvault(self) -> Keyvault:
        return Keyvault(config.get("service_connections", "keyvault"))

    @provider
    @singleton
    def provide_stid_service(self, request_handler: RequestHandler) -> StidService:
        return StidService(request_handler=request_handler)


class MqttModule(Module):
    @provider
    @singleton
    def provide_mqtt_client(self) -> MqttClientInterface:
        mqtt_enabled: bool = config.getboolean("modules", "mqtt_enabled")
        if mqtt_enabled:
            return MqttClient()
        return None


modules: dict = {
    "api": {"default": APIModule},
    "authentication": {"default": AuthenticationModule},
    "queues": {"default": QueuesModule},
    "request_handler": {"default": RequestHandlerModule},
    "robot": {"default": RobotModule},
    "mission_planner": {
        "default": LocalPlannerModule,
        "local": LocalPlannerModule,
        "echo": EchoPlannerModule,
    },
    "service": {"default": ServiceModule},
    "state_machine": {"default": StateMachineModule},
    "storage": {
        "default": LocalStorageModule,
        "local": LocalStorageModule,
        "blob": BlobStorageModule,
        "slimm": SlimmStorageModule,
    },
    "mqtt_enabled": {
        "default": MqttModule,
        "false": MqttModule,
        "true": MqttModule,
    },
    "utilities": {"default": UtilitiesModule},
}


def get_injector_modules() -> Tuple[List[Module], List[str]]:
    injector_modules: List[Module] = []
    module_config_keys: List[str] = []

    module_config: dict = dict(config.items("modules"))

    for module_key, module in modules.items():
        if module_key not in module_config:
            injector_modules.append(module["default"])
            module_config_keys.append(f"{module_key} : default")

        else:
            config_list: List[str] = [
                x.strip() for x in module_config[module_key].split(",")
            ]

            for module_config_key in config_list:
                injector_modules.append(module[module_config_key])
                module_config_keys.append(f"{module_key} : {module_config_key}")

    return injector_modules, module_config_keys



import shutil
import time
from copy import deepcopy
from datetime import datetime, timedelta
from pathlib import Path
from typing import List

import numpy as np
import pytest
from fastapi import FastAPI
from injector import Injector
from starlette.testclient import TestClient

from isar.apis.api import API
from isar.config import config
from isar.models.mission import Mission
from isar.modules import (
    APIModule,
    LocalPlannerModule,
    LocalStorageModule,
    QueuesModule,
    RequestHandlerModule,
    RobotModule,
    ServiceModule,
    StateMachineModule,
    UtilitiesModule,
)
from isar.services.readers.base_reader import BaseReader
from isar.state_machine.states_enum import States
from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.position import Position
from robot_interface.models.mission import DriveToPose
from tests.isar.state_machine.test_state_machine import StateMachineThread
from tests.test_modules import MockNoAuthenticationModule


@pytest.fixture()
def injector_turtlebot():
    config.set("DEFAULT", "robot_package", "isar_turtlebot")
    config.set("DEFAULT", "local_storage_path", "./tests/results")
    config.set(
        "DEFAULT",
        "predefined_missions_folder",
        "./tests/integration/turtlebot/config/missions",
    )

    config.set("DEFAULT", "maps_folder", "tests/integration/turtlebot/config/maps")
    config.set("DEFAULT", "default_map", "turtleworld")

    return Injector(
        [
            APIModule,
            MockNoAuthenticationModule,
            CoordinateModule,
            QueuesModule,
            ReaderModule,
            RequestHandlerModule,
            RobotModule,
            ServiceModule,
            StateMachineModule,
            LocalPlannerModule,
            LocalStorageModule,
            UtilitiesModule,
        ]
    )


@pytest.fixture()
def state_machine_thread(injector_turtlebot) -> StateMachineThread:
    return StateMachineThread(injector=injector_turtlebot)


@pytest.fixture(autouse=True)
def run_before_and_after_tests():
    results_folder: Path = Path("tests/results")
    yield

    print("Removing temporary results folder for testing")
    if results_folder.exists():
        shutil.rmtree(results_folder)
    print("Cleanup finished")


def test_successful_mission(
    injector_turtlebot, state_machine_thread, access_token
) -> None:
    integration_test_timeout: timedelta = timedelta(minutes=5)
    app: FastAPI = injector_turtlebot.get(API).get_app()
    client: TestClient = TestClient(app=app)

    _id: int = 2

    response = client.post(
        f"schedule/start-mission?ID={_id}",
        headers={"Authorization": "Bearer {}".format(access_token)},
    )
    time.sleep(5)

    mission_id: str = state_machine_thread.state_machine.current_mission.id
    mission: Mission = deepcopy(state_machine_thread.state_machine.current_mission)

    start_time: datetime = datetime.utcnow()
    while state_machine_thread.state_machine.current_state != States.Idle:
        if (datetime.utcnow() - start_time) > integration_test_timeout:
            raise TimeoutError
        time.sleep(5)

    mission_result_folder: Path = Path(f"tests/results/{mission_id}")
    image_folder: Path = mission_result_folder.joinpath("sensor_data/image")
    image_navi_file: Path = image_folder.joinpath(f"{mission_id}_image_NAVI.json")

    thermal_folder: Path = mission_result_folder.joinpath("sensor_data/thermal")
    thermal_navi_file: Path = thermal_folder.joinpath(f"{mission_id}_thermal_NAVI.json")

    mission_metadata_file: Path = mission_result_folder.joinpath(
        f"{mission_id}_META.json"
    )

    image_navi: dict = BaseReader.read_json(image_navi_file)
    position_1: Position = Position.from_list(
        image_navi[0]["position"], frame=Frame.Asset
    )
    position_2: Position = Position.from_list(
        image_navi[1]["position"], frame=Frame.Asset
    )

    thermal_navi: dict = BaseReader.read_json(thermal_navi_file)
    position_3: Position = Position.from_list(
        thermal_navi[0]["position"], frame=Frame.Asset
    )

    mission_metadata: dict = BaseReader.read_json(mission_metadata_file)
    folder_locations: list = mission_metadata["data"]

    actual_positions: list = [position_1, position_2, position_3]
    drive_to_tasks: List[DriveToPose] = [
        task for task in mission.tasks if isinstance(task, DriveToPose)
    ]
    expected_positions: list = [task.pose.position for task in drive_to_tasks]

    assert image_navi_file.exists()
    assert thermal_navi_file.exists()
    assert mission_metadata_file.exists()

    assert len(list(image_folder.glob("*"))) == 3  # Expected two images + metadata
    assert len(list(thermal_folder.glob("*"))) == 2  # Expected one thermal + metadata
    assert len(folder_locations) == 2  # Folders for images and thermal images

    for actual, expected in zip(actual_positions, expected_positions):
        assert np.allclose(actual.to_list(), expected.to_list(), atol=0.2)



import logging
from logging import Logger
from threading import Thread
from typing import List

from injector import Injector

from isar.apis.api import API
from isar.config import config
from isar.config.log import setup_logger
from isar.models.communication.queues.queues import Queues
from isar.modules import get_injector_modules
from isar.state_machine.state_machine import main
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader

if __name__ == "__main__":
    setup_logger()
    logger: Logger = logging.getLogger("main")

    injector_modules, module_config_keys = get_injector_modules()
    injector: Injector = Injector(injector_modules)

    module_config_log = "\n".join(module_config_keys)
    logger.info(f"Loaded the following module configurations:\n{module_config_log}")

    threads: List[Thread] = []

    state_machine_thread: Thread = Thread(
        target=main, name="ISAR State Machine", args=[injector], daemon=True
    )
    threads.append(state_machine_thread)

    uploader: Uploader = Uploader(
        upload_queue=injector.get(Queues).upload_queue,
        storage_handlers=injector.get(List[StorageInterface]),
    )

    uploader_thread: Thread = Thread(
        target=uploader.run, name="ISAR Uploader", daemon=True
    )
    threads.append(uploader_thread)

    host: str = config.get("DEFAULT", "api_host")
    port: int = config.getint("DEFAULT", "api_port")

    api: API = injector.get(API)
    api_thread: Thread = Thread(target=api.run_app, name="ISAR API", daemon=True)
    threads.append(api_thread)

    for thread in threads:
        thread.start()
        logger.info(f"Started thread: {thread.name}")

    for thread in threads:
        thread.join()
        logger.info(f"Joined thread: {thread.name}")



import pytest

from isar.models.mission import Mission
from isar.state_machine.state_machine import StateMachine, States
from robot_interface.models.mission import DriveToPose
from tests.mocks.robot_variables import mock_pose


@pytest.mark.parametrize(
    "mock_should_start_mission, next_expected_state",
    [
        (
            (
                True,
                Mission([DriveToPose(pose=mock_pose())]),
            ),
            States.Send,
        ),
        ((False, None), States.Idle),
    ],
)
def test_idle_iterate(
    idle_state,
    mocker,
    mock_should_start_mission,
    next_expected_state,
):
    mocker.patch.object(
        StateMachine, "should_start_mission", return_value=mock_should_start_mission
    )
    next_state: States = idle_state._get_next_state()
    assert next_state == next_expected_state



import logging
from http import HTTPStatus

from fastapi import Query, Response
from injector import inject

from isar.apis.models import StartResponse
from isar.mission_planner.mission_planner_interface import (
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.models.communication.messages.start_message import StartMissionMessages
from isar.models.mission import Mission
from isar.services.utilities.scheduling_utilities import SchedulingUtilities


class StartMission:
    @inject
    def __init__(
        self,
        mission_planner: MissionPlannerInterface,
        scheduling_utilities: SchedulingUtilities,
    ):
        self.logger = logging.getLogger("api")
        self.mission_planner = mission_planner
        self.scheduling_utilities = scheduling_utilities

    def post(
        self,
        response: Response,
        mission_id: int = Query(
            ...,
            alias="ID",
            title="Mission ID",
            description="ID-number for predefined mission",
        ),
    ):
        self.logger.info("Received request to start new mission")

        try:
            mission: Mission = self.mission_planner.get_mission(mission_id)
        except MissionPlannerError as e:
            message = StartMissionMessages.mission_not_found()
            self.logger.error(e)
            response.status_code = HTTPStatus.NOT_FOUND.value
            return StartResponse(message=message.message, started=message.started)

        ready, response_ready_start = self.scheduling_utilities.ready_to_start_mission()
        if not ready:
            message, status_code_ready_start = response_ready_start
            response.status_code = status_code_ready_start.value
            return StartResponse(message=message.message, started=message.started)

        self.logger.info(f"Starting mission: {mission.id}")
        response_scheduler = self.scheduling_utilities.start_mission(mission=mission)
        self.logger.info(f"Received response from State Machine: {response_scheduler}")

        message_scheduler, status_code_scheduler = response_scheduler
        response.status_code = status_code_scheduler.value
        return StartResponse(
            message=message_scheduler.message,
            started=message_scheduler.started,
            mission_id=str(mission.id),
        )



import logging
from typing import Optional

from alitra import AlignFrames, FrameTransform, MapConfig, Point, Quaternion

from isar.services.coordinates.coordinate_utilities import (
    orientation_from_quaternion,
    quaternion_from_orientation,
)
from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.orientation import Orientation
from robot_interface.models.geometry.pose import Pose
from robot_interface.models.geometry.position import Position


class Transformation:
    def __init__(self, map_config: MapConfig):
        self.transform: FrameTransform = AlignFrames.align_frames(
            map_config.robot_reference_points, map_config.asset_reference_points, "z"
        )

    def transform_position(self, position: Position, to_: Frame) -> Optional[Position]:
        """
        Transforms a Position object to a Position object in the to_ frame.
        :param position: A Position object.
        :param to_: Coordinate frame to transfer to given as a Frame object.
        :return: A Position object with to_ as the frame.
        """
        points_from: Point = Point(
            x=position.x, y=position.y, z=position.z, frame=position.frame.value
        )
        point_to: Point = self.transform.transform_point(
            points_from, from_=position.frame.value, to_=to_.value
        )

        position_to: Position = Position(
            x=point_to.x, y=point_to.y, z=point_to.z, frame=to_
        )

        return position_to

    def transform_orientation(
        self, orientation: Orientation, to_: Frame
    ) -> Optional[Orientation]:
        """
        Transforms an Orientation object to an Orientation object in the to_ frame.
        :param orientation: An Orientation object.
        :param to_: Coordinate frame to transfer to given as a Frame object.
        :return: An Orientation object with to_ as the frame.
        """
        quaternion_from: Quaternion = quaternion_from_orientation(
            orientation=orientation
        )

        quaternion_to: Quaternion = self.transform.transform_quaternion(
            quaternion_from, from_=orientation.frame.value, to_=to_.value
        )

        orientation_to: Orientation = orientation_from_quaternion(
            quaternion=quaternion_to
        )
        return orientation_to

    def transform_pose(self, pose: Pose, to_: Frame) -> Optional[Pose]:
        """
        Transforms a Pose object to a Pose object in the to_ frame.
        :param pose: A Pose object.
        :param to_: Coordinate frame to transfer to given as a Frame object.
        :return: A Pose object with to_ as the frame.
        """
        position_to: Position = self.transform_position(position=pose.position, to_=to_)
        orientation_to: Orientation = self.transform_orientation(
            orientation=pose.orientation, to_=to_
        )

        pose_to: Pose = Pose(
            position=position_to, orientation=orientation_to, frame=to_
        )

        return pose_to



from http import HTTPStatus

import pytest

from isar.models.communication.messages import StartMissionMessages
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.state_machine.states_enum import States
from tests.mocks.status import mock_mission_definition, mock_status


class TestSchedulingUtilities:
    @pytest.mark.parametrize(
        "mock_return, expected_ready",
        [
            (mock_status(mission_in_progress=False, current_state=States.Idle), True),
            (mock_status(mission_in_progress=True, current_state=States.Idle), False),
            (
                mock_status(mission_in_progress=False, current_state=States.Monitor),
                False,
            ),
            (mock_status(mission_in_progress=False, current_state=States.Off), False),
            (
                mock_status(mission_in_progress=False, current_state=States.Finalize),
                False,
            ),
            (mock_status(mission_in_progress=False, current_state=States.Send), False),
            (
                mock_status(mission_in_progress=True, current_state=States.Monitor),
                False,
            ),
            (mock_status(mission_in_progress=True, current_state=States.Off), False),
            (
                mock_status(mission_in_progress=True, current_state=States.Finalize),
                False,
            ),
            (mock_status(mission_in_progress=True, current_state=States.Send), False),
        ],
    )
    def test_ready_to_start_mission(
        self, mocker, scheduling_utilities, mock_return, expected_ready
    ):
        mocker.patch.object(QueueUtilities, "check_queue", return_value=mock_return)
        ready, response = scheduling_utilities.ready_to_start_mission()
        assert ready == expected_ready

    @pytest.mark.parametrize(
        "mock_return, mission, expected_output",
        [
            (
                [StartMissionMessages.success()],
                mock_mission_definition(),
                HTTPStatus.OK,
            ),
            (
                [QueueTimeoutError()],
                mock_mission_definition(),
                HTTPStatus.REQUEST_TIMEOUT,
            ),
        ],
    )
    def test_start_mission(
        self, mocker, scheduling_utilities, mock_return, mission, expected_output
    ):
        mocker.patch.object(QueueUtilities, "check_queue", side_effect=mock_return)

        message, status_code = scheduling_utilities.start_mission(mission)

        assert status_code == expected_output



from queue import Queue

from isar.models.communication.queues.queue_io import QueueIO


class Queues:
    def __init__(self):
        self.start_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.stop_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.mission_status: QueueIO = QueueIO()
        self.single_action: QueueIO = QueueIO(input_size=1, output_size=1)
        self.upload_queue: Queue = Queue(maxsize=10)



from abc import ABC
from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional
from uuid import UUID, uuid4

from robot_interface.models.geometry.pose import Pose


@dataclass
class TimeIndexedPose:
    pose: Pose
    time: datetime


@dataclass
class InspectionMetadata(ABC):
    start_time: datetime
    time_indexed_pose: TimeIndexedPose
    file_type: str
    tag_id: Optional[str] = field(default=None, init=False)
    additional: Optional[dict] = field(default=None, init=False)


@dataclass
class ImageMetadata(InspectionMetadata):
    pass


@dataclass
class ThermalImageMetadata(InspectionMetadata):
    pass


@dataclass
class Inspection:
    id: UUID = field(default_factory=uuid4, init=False)
    metadata: InspectionMetadata
    data: Optional[bytes] = field(default=None, init=False)


@dataclass
class Image(Inspection):
    metadata: ImageMetadata


@dataclass
class ThermalImage(Inspection):
    metadata: ThermalImageMetadata



from datetime import datetime

from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.orientation import Orientation
from robot_interface.models.geometry.pose import Pose
from robot_interface.models.geometry.position import Position
from robot_interface.models.inspection.inspection import (
    ImageMetadata,
    TimeIndexedPose,
)

MISSION_ID = "some-mission-id"
ARBITRARY_IMAGE_METADATA = ImageMetadata(
    datetime.now(),
    TimeIndexedPose(
        Pose(
            Position(0, 0, 0, Frame.Asset),
            Orientation(0, 0, 0, 1, Frame.Asset),
            Frame.Asset,
        ),
        datetime.now(),
    ),
    file_type="jpg",
)
DATA_BYTES: bytes = b"Lets say this is some image data"



from dataclasses import asdict
from pathlib import Path
from typing import Any

import pytest

from isar.models.mission import Mission
from isar.services.readers.base_reader import BaseReader
from robot_interface.models.geometry.pose import Pose
from robot_interface.models.mission import Task
from tests.mocks.mission_definition import mock_mission_definition
from tests.mocks.robot_variables import mock_pose
from tests.mocks.task import MockTask
from tests.utilities import Utilities


class TestBaseReader:
    @pytest.mark.parametrize(
        "location, expected_output",
        [
            (Path("./tests/test_data/test_mission_nofile.json"), None),
            (Path("./tests/test_data/test_mission_working_notasks.json"), dict),
            (Path("./tests/test_data/test_mission_working.json"), dict),
            (Path("./tests/test_data/test_mission_not_working.json"), dict),
            (Path("./tests/test_data/test_json_file.json"), list),
        ],
    )
    def test_read_json(self, location, expected_output):
        try:
            content = BaseReader.read_json(location)
        except Exception:
            content = None
        assert Utilities.compare_two_arguments(content, expected_output)

    @pytest.mark.parametrize(
        "dataclass_dict, expected_dataclass",
        [
            (asdict(mock_mission_definition("long_mission")), Mission),
            (asdict(MockTask.drive_to()), Task),
            (asdict(MockTask.take_image_in_coordinate_direction()), Task),
            (asdict(mock_pose()), Pose),
        ],
    )
    def test_dict_to_dataclass(self, dataclass_dict: dict, expected_dataclass: Any):
        content = BaseReader.dict_to_dataclass(dataclass_dict, expected_dataclass)
        assert type(content) is expected_dataclass



import pytest

from robot_interface.models.mission import Task, TaskStatus
from tests.mocks.task import MockTask


@pytest.mark.parametrize(
    "mock_status, expected_output",
    [
        (TaskStatus.Completed, True),
        (TaskStatus.Completed, True),
        (TaskStatus.Unexpected, False),
        (TaskStatus.Failed, True),
    ],
)
def test_task_completed(monitor, mock_status, expected_output):
    task: Task = MockTask.drive_to()
    task.status = mock_status
    task_completed: bool = monitor._task_completed(
        task=task,
    )

    assert task_completed == expected_output



from dataclasses import dataclass
from typing import List

import numpy as np
from alitra import Quaternion
from alitra.convert import quaternion_to_euler

from robot_interface.models.geometry.frame import Frame


@dataclass
class Orientation:
    """
    This class represents an orientation using quaternions. The quaternion is used throughout the project.
    Methods that utilize Euler angles will all follow the yaw, pitch, roll convention which rotates around the ZYX axis
    with intrinsic rotations.
    """

    x: float
    y: float
    z: float
    w: float
    frame: Frame

    def __eq__(self, other):
        if not isinstance(other, Orientation):
            return False
        if (
            np.allclose(
                a=[self.x, self.y, self.z, self.w],
                b=[other.x, other.y, other.z, other.w],
                atol=1e-10,
            )
            and self.frame == other.frame
        ):
            return True
        return False

    def as_euler(
        self,
        degrees: bool = False,
        wrap_angles: bool = False,
    ) -> list:
        """
        Retrieve the orientation as yaw, pitch, roll Euler coordinates. This function uses the ZYX intrinsic rotations
        as standard convention.
        :param degrees: Set to true to retrieve angles as degrees.
        :return: List of euler angles [yaw, pitch, roll]
        """

        euler: list = (
            quaternion_to_euler(
                Quaternion(
                    x=self.x, y=self.y, z=self.z, w=self.w, frame=self.frame.value
                ),
                sequence="ZYX",
                degrees=degrees,
            )
            .as_np_array()
            .tolist()
        )

        if wrap_angles:
            base = 360.0 if degrees else 2 * np.pi
            euler = list(map(lambda angle: ((angle + base) % (base)), euler))

        return euler

    def yaw(self, degrees: bool = False, wrap_angles: bool = False) -> float:
        euler_angles: list = self.as_euler(degrees=degrees, wrap_angles=wrap_angles)
        return euler_angles[0]

    def to_list(self) -> List[float]:
        return [self.x, self.y, self.z, self.w]



from dataclasses import dataclass
from datetime import date, datetime
from typing import Optional, Union
from uuid import UUID

from isar.config import config

additional_meta: dict = {}


@dataclass
class MissionMetadata:
    mission_id: Union[UUID, int, str, None]
    coordinate_reference_system: str = config.get(
        "metadata", "coordinate_reference_system"
    )
    vertical_reference_system: str = config.get("metadata", "vertical_reference_system")
    data_classification: str = config.get("metadata", "data_classification")
    source_url: Optional[str] = None
    plant_code: str = config.get("metadata", "plant_code")
    plant_name: str = config.get("metadata", "plant_name")
    media_orientation_reference_system: str = config.get(
        "metadata", "media_orientation_reference_system"
    )
    robot_id: str = config.get("metadata", "robot_type")
    mission_date: date = datetime.utcnow().date()



from alitra import Quaternion

from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.orientation import Orientation


def orientation_from_quaternion(quaternion: Quaternion) -> Orientation:
    frame: Frame
    if quaternion.frame == Frame.Robot.value:
        frame = Frame.Robot
    elif quaternion.frame == Frame.Asset.value:
        frame = Frame.Asset
    else:
        raise ValueError(f"Frame: {quaternion.frame} is not supported.")

    return Orientation(
        x=quaternion.x, y=quaternion.y, z=quaternion.z, w=quaternion.w, frame=frame
    )


def quaternion_from_orientation(orientation: Orientation) -> Quaternion:
    return Quaternion(
        x=orientation.x,
        y=orientation.y,
        z=orientation.z,
        w=orientation.w,
        frame=orientation.frame.value,
    )



import logging
from http import HTTPStatus

from fastapi import Response
from injector import inject

from isar.apis.models import StopResponse
from isar.config import config
from isar.models.communication.messages import StopMessage, StopMissionMessages
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.models.communication.queues.queues import Queues
from isar.services.utilities.queue_utilities import QueueUtilities


class StopMission:
    @inject
    def __init__(self, queues: Queues):
        self.logger = logging.getLogger("api")
        self.queues = queues
        self.queue_timeout: int = config.getint("DEFAULT", "queue_timeout")

    def post(self, response: Response):
        self.logger.info("Received request to stop current mission")
        self.queues.stop_mission.input.put(True)

        try:
            message: StopMessage = QueueUtilities.check_queue(
                self.queues.stop_mission.output,
                self.queue_timeout,
            )
        except QueueTimeoutError:
            message = StopMissionMessages.queue_timeout()
            self.logger.error((message, HTTPStatus.REQUEST_TIMEOUT))
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value
            return StopResponse(message=message.message, stopped=message.stopped)
        self.logger.info(response)
        return StopResponse(message=message.message, stopped=message.stopped)



from dataclasses import dataclass

from isar.models.communication.message import Message


@dataclass
class StopMessage(Message):
    stopped: bool


class StopMissionMessages:
    @staticmethod
    def success() -> StopMessage:
        return StopMessage(stopped=True, message="Mission stopping")

    @staticmethod
    def queue_timeout() -> StopMessage:
        return StopMessage(
            stopped=False, message="Waiting for return message on queue timed out"
        )

    @staticmethod
    def no_active_missions() -> StopMessage:
        return StopMessage(stopped=False, message="No mission is currently active")



from enum import Enum


class States(str, Enum):

    Off = "off"
    Idle = "idle"
    Send = "send"
    Monitor = "monitor"
    Finalize = "finalize"

    def __repr__(self):
        return self.value



import json
from dataclasses import asdict, is_dataclass
from datetime import date, datetime
from types import SimpleNamespace
from uuid import UUID

import numpy as np

from robot_interface.models.geometry.orientation import Orientation


class JsonService:
    """
    Contains helper functions and custom encoders for handling Json objects.
    """

    @staticmethod
    def to_object(json_string: str) -> SimpleNamespace:
        return json.loads(json_string, object_hook=lambda d: SimpleNamespace(**d))


class EnhancedJSONEncoder(json.JSONEncoder):
    """
    Custom JSONEncoder used in this project. Of special note is the ability to encode dataclasses.
    """

    def default(self, o):
        if is_dataclass(o):
            return asdict(o)
        if isinstance(o, UUID):
            return str(o)
        if isinstance(o, Orientation):
            return o.__dict__
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, date):
            return o.isoformat()
        if isinstance(o, bytes):
            return "<<non-serializable: bytes>>"
        return super().default(o)



import time
from collections import deque
from pathlib import Path
from threading import Thread
from typing import List

import pytest
from injector import Injector

from isar.mission_planner.local_planner import LocalPlanner
from isar.models.communication.queues.queues import Queues
from isar.models.mission import Mission
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine, main
from isar.state_machine.states_enum import States
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader
from robot_interface.models.mission import DriveToPose, TakeImage, Task
from robot_interface.models.mission.status import TaskStatus
from tests.mocks.mission_definition import default_mission
from tests.mocks.robot_interface import MockRobot
from tests.mocks.robot_variables import mock_pose
from tests.mocks.task import MockTask


class StateMachineThread(object):
    def __init__(self, injector) -> None:
        self.injector: Injector = injector
        self.state_machine: StateMachine = injector.get(StateMachine)
        self._thread: Thread = Thread(target=main, args=[injector])
        self._thread.daemon = True
        self._thread.start()


class UploaderThread(object):
    def __init__(self, injector) -> None:
        self.injector: Injector = injector
        self.uploader: Uploader = Uploader(
            upload_queue=self.injector.get(Queues).upload_queue,
            storage_handlers=injector.get(List[StorageInterface]),
        )
        self._thread: Thread = Thread(target=self.uploader.run)
        self._thread.daemon = True
        self._thread.start()


@pytest.fixture
def state_machine_thread(injector) -> StateMachineThread:
    return StateMachineThread(injector)


@pytest.fixture
def uploader_thread(injector) -> UploaderThread:
    return UploaderThread(injector=injector)


def get_mission():
    mission_reader: LocalPlanner = LocalPlanner()
    mission: Mission = mission_reader.read_mission_from_file(
        Path("./tests/test_data/test_mission_working.json")
    )
    return mission


def test_initial_off(state_machine):
    assert state_machine.state == "off"


@pytest.mark.parametrize("should_send, expected_output", [(True, True), (False, False)])
def test_should_send_status(state_machine, should_send, expected_output):
    if should_send is not None:
        state_machine.queues.mission_status.input.put(should_send)
    send: bool = state_machine.should_send_status()

    assert send is expected_output


def test_send_status(state_machine):
    state_machine.send_status()
    message = state_machine.queues.mission_status.output.get()
    assert message


def test_reset_state_machine(state_machine):
    next_state = state_machine.reset_state_machine()

    assert not state_machine.mission_in_progress
    assert state_machine.current_task is None
    assert state_machine.current_mission.tasks == []
    assert next_state is States.Idle


empty_mission: Mission = Mission([], None)


@pytest.mark.parametrize(
    "mission, mission_in_progress, expected_output",
    [
        (None, True, (False, None)),
        (empty_mission, True, (False, None)),
        (empty_mission, False, (True, empty_mission)),
    ],
)
def test_should_start_mission(
    state_machine, mission, mission_in_progress, expected_output
):
    state_machine.queues.start_mission.input.put(mission)
    state_machine.mission_in_progress = mission_in_progress
    output = state_machine.should_start_mission()

    assert output == expected_output


def test_start_mission(state_machine):
    mission: Mission = default_mission
    state_machine.start_mission(mission=mission)
    message = state_machine.queues.start_mission.output.get()
    assert state_machine.mission_in_progress
    assert message


@pytest.mark.parametrize(
    "should_stop, mission_in_progress, expected_output",
    [
        (True, True, True),
        (True, False, False),
        (False, True, False),
        (False, False, False),
        (None, False, False),
        (None, True, False),
    ],
)
def test_should_stop_mission(
    state_machine, should_stop, mission_in_progress, expected_output
):
    if should_stop is not None:
        state_machine.queues.stop_mission.input.put(should_stop)

    state_machine.mission_in_progress = mission_in_progress
    start: bool = state_machine.should_stop_mission()

    assert start is expected_output


def test_stop_mission(state_machine):
    state_machine.stop_mission()
    message = state_machine.queues.stop_mission.output.get()
    assert not state_machine.mission_in_progress
    assert message


def test_state_machine_transitions(injector, state_machine_thread):
    task: Task = DriveToPose(pose=mock_pose())
    mission: Mission = Mission([task])

    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    message, _ = scheduling_utilities.start_mission(mission=mission)
    assert message.started

    time.sleep(1)
    expected_transitions_list = deque(
        [
            States.Idle,
            States.Send,
            States.Monitor,
            States.Send,
            States.Finalize,
            States.Idle,
        ]
    )
    assert (
        state_machine_thread.state_machine.transitions_list == expected_transitions_list
    )


def test_state_machine_failed_dependency(injector, state_machine_thread, mocker):
    driveto_task: Task = DriveToPose(pose=mock_pose())
    inspection_task: Task = MockTask.take_image_in_coordinate_direction()
    mission: Mission = Mission([driveto_task, inspection_task])
    mission.set_task_dependencies()

    mocker.patch.object(MockRobot, "task_status", return_value=TaskStatus.Failed)

    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    message, _ = scheduling_utilities.start_mission(mission=mission)
    assert message.started

    time.sleep(1)
    expected_transitions_list = deque(
        [
            States.Idle,
            States.Send,
            States.Monitor,
            States.Send,
            States.Finalize,
            States.Idle,
        ]
    )
    assert (
        state_machine_thread.state_machine.transitions_list == expected_transitions_list
    )


def test_state_machine_with_successful_collection(
    injector, state_machine_thread, uploader_thread
):
    storage_mock: StorageInterface = injector.get(List[StorageInterface])[0]

    task: TakeImage = MockTask.take_image_in_coordinate_direction()
    mission: Mission = Mission([task])
    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)

    message, _ = scheduling_utilities.start_mission(mission=mission)
    assert message.started
    time.sleep(1)
    expected_transitions_list = deque(
        [
            States.Idle,
            States.Send,
            States.Monitor,
            States.Send,
            States.Finalize,
            States.Idle,
        ]
    )
    expected_stored_items = 1
    assert len(storage_mock.stored_inspections) == expected_stored_items
    assert (
        state_machine_thread.state_machine.transitions_list == expected_transitions_list
    )


def test_state_machine_with_unsuccessful_collection(
    injector, mocker, state_machine_thread
):
    storage_mock: StorageInterface = injector.get(List[StorageInterface])[0]

    mocker.patch.object(MockRobot, "get_inspections", return_value=[])

    task: TakeImage = MockTask.take_image_in_coordinate_direction()
    mission: Mission = Mission([task])
    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)

    message, _ = scheduling_utilities.start_mission(mission=mission)
    assert message.started
    time.sleep(1)
    expected_transitions_list = deque(
        [
            States.Idle,
            States.Send,
            States.Monitor,
            States.Send,
            States.Finalize,
            States.Idle,
        ]
    )
    expected_stored_items = 0
    assert len(storage_mock.stored_inspections) == expected_stored_items
    print(state_machine_thread.state_machine.transitions_list)
    assert (
        state_machine_thread.state_machine.transitions_list == expected_transitions_list
    )



from abc import ABCMeta, abstractmethod
from typing import Sequence

from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission import InspectionTask, Task, TaskStatus


class RobotInterface(metaclass=ABCMeta):
    """Interface to communicate with robots."""

    @abstractmethod
    def initiate_task(self, task: Task) -> None:
        """Send a task to the robot and start the execution of the task

        Parameters
        ----------
        task : Task
            The task that should be initiated on the robot.

        Returns
        -------
        None

        Raises
        ------
        RobotException
            If the task is not initiated.

        """
        raise NotImplementedError

    @abstractmethod
    def task_status(self) -> TaskStatus:
        """Gets the status of the currently active task on robot.

        Parameters
        ----------
        None

        Returns
        -------
        TaskStatus
            Status of the execution of current task.

        Raises:
        ------
        RobotException
            If the task status can't be retrieved.

        """
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """Stops the execution of the current task and stops the movement of the robot.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        RobotException
            If the robot is not stopped.

        """
        raise NotImplementedError

    @abstractmethod
    def get_inspections(self, task: InspectionTask) -> Sequence[Inspection]:
        """Return the inspections connected to the given task.

        Parameters
        ----------
        task : Task

        Returns
        -------
        Sequence[InspectionResult]
            List containing all the inspection results connected to the given task.

        Raises
        ------
        RobotException
            If the inspection results can't be retrieved.

        """
        raise NotImplementedError



import logging
from typing import TYPE_CHECKING

from injector import inject
from transitions import State

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Finalize(State):
    @inject
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="finalize", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")

    def start(self):
        self.state_machine.update_state()

        self.state_machine.log_task_overview(mission=self.state_machine.current_mission)
        next_state = self.state_machine.reset_state_machine()
        self.state_machine.to_next_state(next_state)

    def stop(self):
        self._log_state_transitions()

    def _log_state_transitions(self):
        state_transitions: str = ", ".join(
            [
                f"\n  {transition}" if (i + 1) % 10 == 0 else f"{transition}"
                for i, transition in enumerate(
                    list(self.state_machine.transitions_list)
                )
            ]
        )

        self.logger.info(f"State transitions:\n  {state_transitions}")



from isar.models.communication.status import Status
from isar.models.mission import Mission
from isar.state_machine.states_enum import States
from robot_interface.models.mission import Task, TaskStatus
from tests.mocks.mission_definition import mock_mission_definition
from tests.mocks.task import MockTask


def mock_status(
    mission_in_progress: bool = True,
    task_status: TaskStatus = TaskStatus.Scheduled,
    current_task: Task = MockTask.take_image_in_coordinate_direction(),
    current_state: States = States.Idle,
    current_mission: Mission = mock_mission_definition(),
) -> Status:

    current_task.status = task_status
    scheduled_status = Status(
        mission_in_progress=mission_in_progress,
        current_task=current_task,
        current_state=current_state,
        current_mission=current_mission,
    )
    return scheduled_status



from dataclasses import dataclass

from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.orientation import Orientation
from robot_interface.models.geometry.position import Position


@dataclass
class Pose:
    position: Position
    orientation: Orientation
    frame: Frame



import logging
from typing import Dict, List, Union

from azure.identity import DefaultAzureCredential
from injector import inject
from requests import Response

from isar.config import config
from isar.config.predefined_poses.predefined_poses import predefined_poses
from isar.mission_planner.mission_planner_interface import (
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.models.mission import Mission
from isar.services.auth.azure_credentials import AzureCredentials
from isar.services.service_connections.request_handler import RequestHandler
from isar.services.service_connections.stid.stid_service import StidService
from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.pose import Pose
from robot_interface.models.geometry.position import Position
from robot_interface.models.mission import DriveToPose, TakeImage, TakeThermalImage


class EchoPlanner(MissionPlannerInterface):
    @inject
    def __init__(
        self,
        request_handler: RequestHandler,
        stid_service: StidService,
    ):
        self.request_handler: RequestHandler = request_handler
        self.stid_service: StidService = stid_service
        self.credentials: DefaultAzureCredential = (
            AzureCredentials.get_azure_credentials()
        )
        self.logger = logging.getLogger("api")

    def get_mission(self, mission_id: int) -> Mission:
        """
        Retrieve robot mission from echo mission planner with specified id.
        :param mission_id: Unique id of echo mission plan
        :return: Mission object
        """
        mission_plan: dict = self._mission_plan(mission_id)

        try:
            plan_items: List[dict] = mission_plan["planItems"]
        except KeyError as e:
            self.logger.error("Echo request body don't contain expected keys")
            raise MissionPlannerError from e

        mission: Mission = Mission(tasks=[])

        for plan_item in plan_items:
            try:
                tag_name: str = plan_item["tag"]
                sensors: List[str] = [
                    sensor_item["sensorTypeKey"]
                    for sensor_item in plan_item["sensorTypes"]
                ]
            except KeyError:
                self.logger.error("Echo request body don't contain expected keys")
                continue

            try:
                drive_task: DriveToPose = self._create_drive_task(tag_name=tag_name)
                inspection_tasks: List[
                    Union[TakeImage, TakeThermalImage]
                ] = self._create_inspection_tasks_from_sensor_types(
                    tag_name=tag_name, sensors=sensors
                )
            except (ValueError, Exception) as e:
                self.logger.error(e)
                continue

            mission.tasks.append(drive_task)
            mission.tasks.extend(inspection_tasks)

        mission.set_task_dependencies()
        return mission

    def _mission_plan(self, mission_id: int) -> dict:
        """
        Get mission plan from echo planner.
        :param mission_id: Unique id of echo mission plan
        """
        client_id: str = config.get("service_connections", "echo_client_id")
        scope: str = config.get("service_connections", "echo_app_scope")
        request_scope: str = f"{client_id}/{scope}"

        token: str = self.credentials.get_token(request_scope).token

        url: str = f"{config.get('service_connections', 'echo_api_url')}/robots/robot-plan/{mission_id}"
        response: Response = self.request_handler.get(
            url=url,
            headers={"Authorization": f"Bearer {token}"},
        )

        return response.json()

    def _create_inspection_tasks_from_sensor_types(
        self, tag_name: str, sensors: List[str]
    ) -> List[Union[TakeImage, TakeThermalImage]]:
        tag_position: Position = self._get_tag_position(tag_name=tag_name)
        inspection_tasks: List[Union[TakeImage, TakeThermalImage]] = []
        for sensor in sensors:
            inspection: Union[
                TakeImage, TakeThermalImage
            ] = self._echo_sensor_type_to_isar_inspection_task(sensor_type=sensor)
            inspection.target = tag_position
            inspection.tag_id = tag_name
            inspection_tasks.append(inspection)
        return inspection_tasks

    def _get_robot_pose(self, tag_name: str) -> Pose:
        """
        Retrieve robot pose corresponding to inspection of a given tag. For now, this is a temporary hard-coded
        solution.
        """
        predefined_pose: Pose = predefined_poses[tag_name]

        if predefined_pose.frame is Frame.Robot:
            raise ValueError(
                f"Frame of predefined pose should be {Frame.Asset} not {Frame.Robot}"
            )

        return predefined_pose

    def _get_tag_position(self, tag_name: str) -> Position:
        tag_position: Position = self.stid_service.tag_position(tag_name)

        return tag_position

    def _create_drive_task(self, tag_name: str) -> DriveToPose:
        robot_pose: Pose = self._get_robot_pose(tag_name=tag_name)

        drive_task: DriveToPose = DriveToPose(pose=robot_pose)
        return drive_task

    @staticmethod
    def _echo_sensor_type_to_isar_inspection_task(
        sensor_type: str,
    ) -> Union[TakeImage, TakeThermalImage]:
        mapping: Dict[str, Union[TakeImage, TakeThermalImage]] = {
            "Picture": TakeImage(target=None),
            "ThermicPicture": TakeThermalImage(target=None),
        }
        return mapping[sensor_type]



import logging
import os
from abc import ABCMeta, abstractmethod
from typing import Any

from paho.mqtt import client as mqtt
from paho.mqtt.client import Client

from isar.config import config


class MqttClientInterface(metaclass=ABCMeta):
    @abstractmethod
    def publish(
        self, topic: str, payload: Any, qos: int = 0, retain: bool = False
    ) -> None:
        """
        Parameters
        ----------
        topic : string
            MQTT topic to publish to
        payload : Any
            Payload to send to publish on the topic
        qos : integer
            Quality of Service
        retain : boolean
            Retain on topic

        Returns
        -------

        """
        pass


class MqttClient(MqttClientInterface):
    def __init__(self):
        self.logger = logging.getLogger("mqtt_client")

        username: str = config.get("service_connections", "mqtt_username")
        password: str = ""
        try:
            password = os.environ["MQTT_PASSWORD"]
        except KeyError:
            self.logger.warning(
                "Failed to retrieve MQTT_PASSWORD from environment. Attempting with empty string as password."
            )

        host: str = config.get("service_connections", "mqtt_host")
        port: int = config.getint("service_connections", "mqtt_port")

        self.client: Client = Client()

        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect

        self.client.username_pw_set(username=username, password=password)
        self.client.connect(host=host, port=port, keepalive=60)

    def on_connect(self, client, userdata, flags, rc):
        self.logger.info("Connection returned result: " + mqtt.connack_string(rc))

    def on_disconnect(self, client, userdata, rc):
        if rc != 0:
            self.logger.warning("Unexpected disconnection from MQTT Broker")

    def publish(self, topic: str, payload: Any, qos: int = 0, retain: bool = False):
        self.client.publish(topic=topic, payload=payload, qos=qos, retain=retain)



from isar.models.communication.queues.queues import Queues


class TestQueues:
    def test_queues(self):
        queues = Queues()
        assert queues.start_mission is not None
        assert (
            queues.start_mission.input is not None
            and queues.start_mission.input.maxsize is 1
        )
        assert (
            queues.start_mission.output is not None
            and queues.start_mission.output.maxsize is 1
        )
        assert queues.stop_mission is not None
        assert (
            queues.stop_mission.input is not None
            and queues.stop_mission.input.maxsize is 1
        )
        assert (
            queues.stop_mission.output is not None
            and queues.stop_mission.output.maxsize is 1
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import List

import numpy as np

from robot_interface.models.geometry.frame import Frame


@dataclass
class Position:
    x: float
    y: float
    z: float
    frame: Frame

    def __eq__(self, other):
        if not isinstance(other, Position):
            return False
        if (
            np.allclose(a=[self.x, self.y, self.z], b=[other.x, other.y, other.z])
            and self.frame == other.frame
        ):
            return True
        return False

    def to_list(self) -> List[float]:
        return [self.x, self.y, self.z]

    @staticmethod
    def from_list(input_list: List, frame: Frame) -> Position:
        return Position(x=input_list[0], y=input_list[1], z=input_list[2], frame=frame)



import json
from typing import Any

import pytest

from isar.models.communication.messages import StartMessage, StopMessage
from isar.services.utilities.json_service import EnhancedJSONEncoder, JsonService
from tests.mocks.status import mock_status


class TestJsonService:
    @pytest.mark.parametrize(
        "original_object",
        [
            StartMessage(message="help", started=False),
            StopMessage(message="Yes, sir!", stopped=True),
        ],
    )
    def test_dataclass_to_object(self, original_object: Any):
        json_string: str = json.dumps(original_object, cls=EnhancedJSONEncoder)
        as_object: Any = JsonService.to_object(json_string)
        assert as_object is not None

    @pytest.mark.parametrize(
        "original_object",
        [mock_status()],
    )
    def test_to_object(self, original_object: Any):
        json_string: str = json.dumps(original_object, cls=EnhancedJSONEncoder)
        as_object: Any = JsonService.to_object(json_string)
        assert as_object is not None



from pathlib import Path

import pytest

from isar.config import config
from isar.models.mission import Mission
from robot_interface.models.mission import TakeThermalImage
from robot_interface.models.mission.task import Task
from tests.utilities import Utilities


@pytest.mark.parametrize(
    "mission_path ,expected_output",
    [
        (
            Path("./tests/test_data/test_mission_working_notasks.json"),
            Mission,
        ),
        (Path("./tests/test_data/test_mission_working.json"), Mission),
    ],
)
def test_get_mission(mission_reader, mission_path, expected_output):
    output: Mission = mission_reader.read_mission_from_file(mission_path)
    assert Utilities.compare_two_arguments(output, expected_output)


@pytest.mark.parametrize(
    "mission_path",
    [
        (Path("./tests/test_data/no_file.json")),
        (Path("./tests/test_data/test_mission_not_working.json")),
    ],
)
def test_get_invalid_mission(mission_reader, mission_path):
    with pytest.raises(Exception):
        mission_reader.read_mission_from_file(mission_path)


@pytest.mark.parametrize(
    "mission_id ,expected_output",
    [(1, Mission)],
)
def test_get_mission_by_id(mission_reader, mission_id, expected_output):
    output = mission_reader.get_mission(mission_id)
    assert Utilities.compare_two_arguments(output, expected_output)


@pytest.mark.parametrize(
    "mission_id",
    [12345, None, config],
)
def test_get_mission_by_invalid_id(mission_reader, mission_id):
    with pytest.raises(Exception):
        mission_reader.get_mission_by_id(mission_id)


def test_valid_predefined_missions_files(mission_reader):
    # Checks that the predefined mission folder contains only valid missions!
    mission_list_dict = mission_reader.get_predefined_missions()
    predefined_mission_folder = Path(
        config.get("DEFAULT", "predefined_missions_folder")
    )
    assert len(list(predefined_mission_folder.glob("*.json"))) == len(
        list(mission_list_dict)
    )
    for file in predefined_mission_folder.glob("*.json"):
        path_to_file = predefined_mission_folder.joinpath(file.name)
        mission: Mission = mission_reader.read_mission_from_file(path_to_file)
        assert mission is not None


def test_thermal_image_task(mission_reader):
    mission_path: Path = Path("./tests/test_data/test_thermal_image_mission.json")
    output: Mission = mission_reader.read_mission_from_file(mission_path)

    task: Task = output.tasks[0]

    assert isinstance(task, TakeThermalImage)
    assert hasattr(task, "target")
    assert task.name == "take_thermal_image"
    assert hasattr(task, "id")
    assert hasattr(task, "tag_id")


def test_mission_dependencies(mission_reader):
    mission_path = Path("./tests/test_data/test_mission_working.json")
    mission: Mission = mission_reader.read_mission_from_file(mission_path)
    mission.set_task_dependencies()

    task_dependencies = [
        None,
        None,
        [mission.tasks[1].id],
        None,
        [mission.tasks[0].id],
        [mission.tasks[1].id, mission.tasks[2].id],
    ]

    for task, dependencies in zip(mission.tasks, task_dependencies):
        assert task.depends_on == dependencies



import logging
import time
from typing import TYPE_CHECKING, List, Sequence, Tuple

from injector import inject
from transitions import State

from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions import RobotException
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission import InspectionTask, Task, TaskStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Monitor(State):
    @inject
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="monitor", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine

        self.logger = logging.getLogger("state_machine")

        self.iteration_counter: int = 0
        self.log_interval = 20

        self.task_status_thread = None

    def start(self):
        self.state_machine.update_state()
        self._run()

    def stop(self):
        if self.state_machine.mqtt_client:
            self.state_machine.publish_task_status()
            self.state_machine.publish_mission()

        self.iteration_counter = 0
        if self.task_status_thread:
            self.task_status_thread.wait_for_thread()
        self.task_status_thread = None

    def _run(self):
        while True:
            if self.state_machine.should_stop_mission():
                self.state_machine.stop_mission()

            if not self.state_machine.mission_in_progress:
                next_state = States.Finalize
                break

            if self.state_machine.should_send_status():
                self.state_machine.send_status()
            if not self.task_status_thread:
                self.task_status_thread = ThreadedRequest(
                    self.state_machine.robot.task_status
                )
                self.task_status_thread.start_thread()

            try:
                task_status: TaskStatus = self.task_status_thread.get_output()
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue
            except RobotException:
                task_status = TaskStatus.Unexpected

            self.state_machine.current_task.status = task_status

            if self._task_completed(task=self.state_machine.current_task):
                if isinstance(self.state_machine.current_task, InspectionTask):
                    self._queue_inspections_for_upload(
                        current_task=self.state_machine.current_task
                    )

                next_state = States.Send
                break
            else:
                self.task_status_thread = None
                time.sleep(self.state_machine.sleep_time)

        self.state_machine.to_next_state(next_state)

    def _queue_inspections_for_upload(self, current_task: InspectionTask):
        inspections: Sequence[Inspection] = self.state_machine.robot.get_inspections(
            task=current_task
        )
        mission_metadata: MissionMetadata = self.state_machine.current_mission.metadata
        for inspection in inspections:
            inspection.metadata.tag_id = current_task.tag_id

            message: Tuple[Inspection, MissionMetadata] = (
                inspection,
                mission_metadata,
            )
            self.state_machine.queues.upload_queue.put(message)
            self.logger.info(f"Inspection: {str(inspection.id)[:8]} queued for upload")

    def _task_completed(self, task: Task) -> bool:
        if task.status == TaskStatus.Unexpected:
            self.logger.error("Task status returned an unexpected status string")
        elif task.status == TaskStatus.Failed:
            self.logger.warning(f"Task: {str(task.id)[:8]} failed")
            return True
        elif task.status == TaskStatus.Completed:
            self.logger.info(
                f"{type(task).__name__} task: {str(task.id)[:8]} completed"
            )
            return True
        return False



from typing import Literal

from isar.models.mission import Mission
from tests.mocks.task import MockTask

default_mission = Mission(
    id="default_mission",
    tasks=[
        MockTask.take_image_in_coordinate_direction(),
        MockTask.drive_to(),
        MockTask.take_image_in_coordinate_direction(),
        MockTask.take_image_in_coordinate_direction(),
    ],
)

long_mission = Mission(
    id="long_mission",
    tasks=[
        MockTask.take_image_in_coordinate_direction(),
        MockTask.take_image_in_coordinate_direction(),
        MockTask.drive_to(),
        MockTask.drive_to(),
        MockTask.take_image_in_coordinate_direction(),
        MockTask.take_image_in_coordinate_direction(),
    ],
)


empty_mission = Mission(id=None, tasks=[])


mission_name_typehints = Literal["default_mission", "empty_mission", "long_mission"]


def mock_mission_definition(
    mission_name: mission_name_typehints = "default_mission",
) -> Mission:
    if mission_name == "default_mission":
        return default_mission
    elif mission_name == "empty_mission":
        return empty_mission
    elif mission_name == "long_mission":
        return long_mission



import inspect
import typing


class Utilities:
    @staticmethod
    def compare_tuples(response_tuple, expected_tuple) -> bool:
        """
        Checks if two tuples are equal. If an expected element is a class
        the function check that the corresponding response is of the same type
        """
        flag = True
        if expected_tuple is None and response_tuple is None:
            return flag
        for value, expected_value in zip(response_tuple, expected_tuple):
            if not Utilities.compare_two_arguments(value, expected_value):
                flag = False
        return flag

    @staticmethod
    def compare_two_arguments(response_argument, expected_argument):
        if inspect.isclass(expected_argument):
            if not isinstance(response_argument, expected_argument):
                return False
        elif hasattr(expected_argument, "__origin__"):
            expected_value_type = expected_argument.__origin__
            expected_value_class = typing.get_args(expected_argument)
            if expected_value_type == list:
                for items in response_argument:
                    if not isinstance(items, expected_value_class):
                        return False
        else:
            if not response_argument == expected_argument:
                return False
        return True



import pytest
from requests import RequestException

from isar.mission_planner.echo_planner import EchoPlanner
from isar.models.mission import Mission
from isar.services.service_connections.stid.stid_service import StidService
from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.position import Position
from robot_interface.models.mission import DriveToPose, TakeImage, TakeThermalImage


@pytest.mark.parametrize(
    "id, mock_return, mock_stid_side_effect, mock_stid, expected_return",
    [
        (
            76,
            {
                "robotPlanId": 76,
                "planItems": [
                    {
                        "planItemId": 1227,
                        "tag": "313-PA-101A",
                        "sensorTypes": [
                            {"sensorTypeKey": "ThermicPicture"},
                            {"sensorTypeKey": "Picture"},
                        ],
                        "sortingOrder": 0,
                        "robotPlanId": 76,
                    },
                ],
            },
            None,
            Position(x=1, y=1, z=0, frame=Frame.Asset),
            {
                "tasks": 3,
                "task_1_type": DriveToPose,
                "task_2_type": TakeThermalImage,
            },
        ),
        (
            76,
            {
                "robotPlanId": 76,
                "planItems": [
                    {
                        "planItemId": 1227,
                        "tag": "355-LD-1003",
                        "sensorTypes": [{"sensorTypeKey": "Picture"}],
                        "sortingOrder": 0,
                        "robotPlanId": 76,
                    },
                ],
            },
            None,
            Position(x=1, y=1, z=0, frame=Frame.Asset),
            {
                "tasks": 2,
                "task_1_type": DriveToPose,
                "task_2_type": TakeImage,
            },
        ),
        (
            76,
            {
                "robotPlanId": 76,
                "planItems": [
                    {
                        "planItemId": 1227,
                        "tag": "wrong_tag",
                        "senorTypes": [{"sensorTypeKey": "Picture"}],
                        "sortingOrder": 0,
                        "robotPlanId": 76,
                    },
                ],
            },
            RequestException,
            Position(x=1, y=1, z=0, frame=Frame.Asset),
            {
                "tasks": 0,
                "task_1_type": None,
                "task_2_type": None,
            },
        ),
    ],
)
def test_get_echo_mission(
    echo_service,
    mocker,
    id,
    mock_return,
    mock_stid_side_effect,
    mock_stid,
    expected_return,
):
    mocker.patch.object(EchoPlanner, "_mission_plan", return_value=mock_return)
    mocker.patch.object(
        StidService,
        "tag_position",
        return_value=mock_stid,
        side_effect=mock_stid_side_effect,
    )
    mission: Mission = echo_service.get_mission(mission_id=id)
    assert len(mission.tasks) == expected_return["tasks"]
    if not len(mission.tasks) == 0:
        assert isinstance(mission.tasks[0], expected_return["task_1_type"])
        assert isinstance(mission.tasks[1], expected_return["task_2_type"])



from enum import Enum


class Frame(str, Enum):
    """
    Robot frame implies the local frame on the operating robot. This will depend on the robots internal mapping.
    Asset frame is the frame of the asset which the robot operates in.
    """

    Robot = "robot"
    Asset = "asset"



from dataclasses import dataclass

from isar.models.communication.message import Message


@dataclass
class StartMessage(Message):
    started: bool


class StartMissionMessages:
    @staticmethod
    def success() -> StartMessage:
        return StartMessage(started=True, message="Mission started")

    @staticmethod
    def mission_not_found() -> StartMessage:
        return StartMessage(started=False, message="Mission was not found")

    @staticmethod
    def mission_in_progress() -> StartMessage:
        return StartMessage(started=False, message="A mission is already in progress")

    @staticmethod
    def queue_timeout() -> StartMessage:
        return StartMessage(
            started=False, message="Waiting for return message on queue timed out"
        )



from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.orientation import Orientation
from robot_interface.models.geometry.pose import Pose
from robot_interface.models.geometry.position import Position


def mock_pose(
    position: Position = Position(x=1, y=1, z=1, frame=Frame.Robot),
    orientation: Orientation = Orientation(
        x=0, y=0, z=0.7071068, w=0.7071068, frame=Frame.Robot
    ),
) -> Pose:
    return Pose(position=position, orientation=orientation, frame=Frame.Robot)



from datetime import datetime
from typing import Optional, Sequence
from uuid import UUID

from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.orientation import Orientation
from robot_interface.models.geometry.pose import Pose
from robot_interface.models.geometry.position import Position
from robot_interface.models.inspection.inspection import (
    Image,
    ImageMetadata,
    Inspection,
    TimeIndexedPose,
)
from robot_interface.models.mission import InspectionTask, Task, TaskStatus
from robot_interface.robot_interface import RobotInterface


class MockRobot(RobotInterface):
    def __init__(
        self,
        initiate_task: bool = True,
        task_status: TaskStatus = TaskStatus.Completed,
        stop: bool = True,
        pose: Pose = Pose(
            position=Position(x=0, y=0, z=0, frame=Frame.Robot),
            orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame.Robot),
            frame=Frame.Robot,
        ),
    ):
        self.initiate_task_return_value: bool = initiate_task
        self.task_status_return_value: TaskStatus = task_status
        self.stop_return_value: bool = stop
        self.robot_pose_return_value: Pose = pose

    def initiate_task(self, task: Task) -> None:
        return

    def task_status(self) -> TaskStatus:
        return self.task_status_return_value

    def stop(self) -> None:
        return

    def get_inspections(self, task: InspectionTask) -> Sequence[Inspection]:
        image: Image = Image(mock_image_metadata())
        image.data = b"Some binary image data"
        return [image]


def mock_image_metadata() -> ImageMetadata:
    return ImageMetadata(
        datetime.now(),
        TimeIndexedPose(
            Pose(
                Position(0, 0, 0, Frame.Robot),
                Orientation(0, 0, 0, 1, Frame.Robot),
                Frame.Robot,
            ),
            datetime.now(),
        ),
        file_type="jpg",
    )



import logging
import time
from typing import TYPE_CHECKING

from transitions import State

from isar.config import config
from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions import RobotException
from robot_interface.models.mission.status import TaskStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Send(State):
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="send", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.send_failure_counter: int = 0
        self.send_failure_counter_limit: int = config.getint(
            "DEFAULT", "send_failure_counter_limit"
        )
        self.logger = logging.getLogger("state_machine")

        self.send_thread = None

    def start(self):
        self.state_machine.update_state()
        self.state_machine.update_current_task()

        if self.state_machine.mqtt_client:
            self.state_machine.publish_task_status()
            self.state_machine.publish_mission()

        self._run()

    def stop(self):
        if self.state_machine.mqtt_client:
            self.state_machine.publish_task_status()
            self.state_machine.publish_mission()

        self.send_failure_counter = 0
        if self.send_thread:
            self.send_thread.wait_for_thread()
        self.send_thread = None

    def _run(self):
        while True:
            if self.state_machine.should_stop_mission():
                self.state_machine.stop_mission()

            if not self.state_machine.mission_in_progress:
                next_state: States = States.Finalize
                break

            if not self.state_machine.current_task:
                next_state: States = States.Finalize
                self.logger.info(
                    f"Completed mission: {self.state_machine.current_mission.id}"
                )
                break

            if self.state_machine.should_send_status():
                self.state_machine.send_status()

            if not self.state_machine._check_dependencies():
                self.state_machine.current_task.status = TaskStatus.Failed
                self.logger.warning(
                    f"Dependency for task {self.state_machine.current_task_index}: "
                    f"{self.state_machine.current_task.name}, not fulfilled, "
                    "skipping to next task"
                )

                next_state: States = States.Send
                break

            if not self.send_thread:
                self.send_thread = ThreadedRequest(
                    self.state_machine.robot.initiate_task
                )
                self.send_thread.start_thread(self.state_machine.current_task)

            try:
                self.send_thread.get_output()
                send_success = True
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue
            except RobotException:
                send_success = False

            if send_success:
                self.state_machine.current_task.status = TaskStatus.Scheduled
                next_state = States.Monitor
                self.logger.info(
                    f"Successfully scheduled {type(self.state_machine.current_task).__name__} "
                    f"task: {str(self.state_machine.current_task.id)[:8]}"
                )
                break
            else:
                self.send_failure_counter += 1
                self.logger.info(f"Sending failed #: {str(self.send_failure_counter)}")
                if self.send_failure_counter >= self.send_failure_counter_limit:
                    self.logger.error(
                        f"Failed to send mission after "
                        f"{self.send_failure_counter_limit} attempts. "
                        f"Cancelling mission."
                    )
                    next_state: States = States.Finalize
                    break
                self.send_thread = None
                time.sleep(self.state_machine.sleep_time)

        self.state_machine.to_next_state(next_state)



import logging
from copy import deepcopy
from http import HTTPStatus
from typing import Optional, Tuple

from injector import inject

from isar.config import config
from isar.models.communication.messages import StartMessage, StartMissionMessages
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.models.communication.queues.queues import Queues
from isar.models.communication.status import Status
from isar.models.mission import Mission
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.state_machine.states_enum import States


class SchedulingUtilities:
    """
    Contains utility functions for scheduling missions from the API. The class handles required thread communication
    through queues to the state machine.
    """

    @inject
    def __init__(
        self,
        queues: Queues,
        queue_timeout: int = config.getint("DEFAULT", "queue_timeout"),
    ):
        self.queues = queues
        self.queue_timeout: int = queue_timeout
        self.logger = logging.getLogger("api")

    def ready_to_start_mission(
        self,
    ) -> Tuple[bool, Optional[Tuple[StartMessage, HTTPStatus]]]:
        """
        Checks the current mission status by communicating with the state machine thread through queues.
        :return: (True, None) if the mission may be started. Otherwise (False, response) with a relevant response
        message indicating the cause.
        """
        self.queues.mission_status.input.put(True)
        try:
            status: Status = QueueUtilities.check_queue(
                self.queues.mission_status.output, self.queue_timeout
            )
        except QueueTimeoutError:
            error_message = (
                StartMissionMessages.queue_timeout(),
                HTTPStatus.REQUEST_TIMEOUT,
            )
            self.logger.error(error_message)
            return False, error_message

        if status.mission_in_progress or status.current_state != States.Idle:
            message = StartMissionMessages.mission_in_progress()
            error_message = message, HTTPStatus.CONFLICT
            return False, error_message
        return True, None

    def start_mission(self, mission: Mission) -> Tuple[StartMessage, HTTPStatus]:
        """
        Starts a mission by communicating with the state machine thread through queues.
        :param mission: A Mission containing the mission tasks to be started.
        :return: (message, status_code) is returned indicating the success and cause of the operation.
        """
        self.queues.start_mission.input.put(deepcopy(mission))
        try:
            start_message: StartMessage = QueueUtilities.check_queue(
                self.queues.start_mission.output, self.queue_timeout
            )
        except QueueTimeoutError:
            error_message = (
                StartMissionMessages.queue_timeout(),
                HTTPStatus.REQUEST_TIMEOUT,
            )
            self.logger.error(error_message)
            return error_message
        if not start_message.started:
            error_message = start_message, HTTPStatus.CONFLICT
            self.logger.error(error_message)
            return error_message

        error_message = start_message, HTTPStatus.OK
        return error_message



from enum import Enum


class TaskStatus(str, Enum):

    NotStarted: str = "not_started"
    Completed: str = "completed"
    Scheduled: str = "scheduled"
    InProgress: str = "in_progress"
    Failed: str = "failed"
    Unexpected: str = "error_unexpected"



import pytest

from isar.state_machine.states_enum import States


@pytest.mark.parametrize(
    "expected_state",
    [States.Idle],
)
def test_reset_state_machine(state_machine, expected_state):
    next_state = state_machine.reset_state_machine()
    assert next_state is expected_state



import logging
from queue import Empty, Queue
from typing import List

from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.storage.storage_interface import StorageException, StorageInterface
from robot_interface.models.inspection.inspection import Inspection


class Uploader:
    def __init__(
        self, upload_queue: Queue, storage_handlers: List[StorageInterface]
    ) -> None:
        self.upload_queue: Queue = upload_queue
        self.storage_handlers: List[StorageInterface] = storage_handlers

        self.logger = logging.getLogger("uploader")

    def run(self) -> None:
        self.logger.info("Started uploader")
        while True:
            inspection: Inspection
            mission_metadata: MissionMetadata
            try:
                inspection, mission_metadata = self.upload_queue.get(timeout=5)
            except Empty:
                continue

            for storage_handler in self.storage_handlers:
                try:
                    storage_handler.store(
                        inspection=inspection, metadata=mission_metadata
                    )
                    self.logger.info(
                        f"Storage handler: {type(storage_handler).__name__} "
                        f"uploaded inspection {str(inspection.id)[:8]}"
                    )
                except StorageException:
                    self.logger.warning(
                        f"Storage handler: {type(storage_handler).__name__} "
                        f"failed to upload inspection: {str(inspection.id)[:8]}"
                    )



import logging
from dataclasses import asdict
from pathlib import Path
from typing import Optional

from injector import inject

from isar.config import config
from isar.mission_planner.mission_planner_interface import (
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.models.mission import Mission
from isar.services.readers.base_reader import BaseReader, BaseReaderError
from robot_interface.models.geometry.frame import Frame

logger = logging.getLogger("api")


class LocalPlanner(MissionPlannerInterface):
    @inject
    def __init__(self):
        self.predefined_mission_folder = Path(
            config.get("DEFAULT", "predefined_missions_folder")
        )

    def get_mission(self, mission_id) -> Mission:
        missions: dict = self.get_predefined_missions()
        if missions is None:
            raise MissionPlannerError("There were no predefined missions")
        try:
            mission: Mission = missions[mission_id]["mission"]
            mission.set_unique_id_and_metadata()
            mission.set_task_dependencies()
            return mission
        except Exception as e:
            raise MissionPlannerError(
                f"Could not get mission : {mission_id} - does not exist {e}"
            ) from e

    @staticmethod
    def read_mission_from_file(mission_path: Path) -> Mission:
        mission_dict: dict = BaseReader.read_json(location=mission_path)
        mission: Mission = BaseReader.dict_to_dataclass(
            dataclass_dict=mission_dict,
            target_dataclass=Mission,
            cast_config=[Frame],
            strict_config=True,
        )

        return mission

    def get_predefined_missions(self) -> dict:
        missions: dict = {}
        invalid_mission_ids: list = []
        json_files = self.predefined_mission_folder.glob("*.json")
        for file in json_files:
            mission_name = file.stem
            path_to_file = self.predefined_mission_folder.joinpath(file.name)
            try:

                mission: Mission = self.read_mission_from_file(path_to_file)
            except BaseReaderError as e:
                logger.warning(
                    f"Failed to read predefined mission {path_to_file} \n {e}"
                )
                continue
            if mission.id in invalid_mission_ids:
                logger.warning(
                    f"Duplicate mission id {mission.id} : {path_to_file.as_posix()}"
                )
            elif mission.id in missions:
                conflicting_file_path = missions[mission.id]["file"]
                logger.warning(
                    f"Duplicate mission id {mission.id} : {path_to_file.as_posix()}"
                    + f" and {conflicting_file_path}"
                )
                invalid_mission_ids.append(mission.id)
                missions.pop(mission.id)
            else:
                missions[mission.id] = {
                    "name": mission_name,
                    "file": path_to_file.as_posix(),
                    "mission": mission,
                }
        return missions

    def list_predefined_missions(self) -> dict:
        missions: dict = self.get_predefined_missions()
        predefined_missions: list = []
        for id, mission in missions.items():
            predefined_missions.append(
                {
                    "id": id,
                    "name": mission["name"],
                    "file": mission["file"],
                    "tasks": asdict(mission["mission"])["tasks"],
                }
            )
        return {"missions": predefined_missions}



from dataclasses import dataclass
from datetime import datetime
from typing import List, Union
from uuid import UUID

from isar.config import config
from isar.models.mission_metadata.mission_metadata import MissionMetadata
from robot_interface.models.mission import TASKS
from robot_interface.models.mission.task import DriveToPose, InspectionTask


@dataclass
class Mission:
    tasks: List[TASKS]
    id: Union[UUID, int, str, None] = None
    metadata: MissionMetadata = None

    def set_unique_id_and_metadata(self) -> None:
        self._set_unique_id()
        self.metadata = MissionMetadata(mission_id=self.id)

    def set_task_dependencies(self):
        last_drive_to_task = None
        for task in self.tasks:
            if isinstance(task, DriveToPose):
                last_drive_to_task = task
                if task.depends_on:
                    task.depends_on = [self.tasks[ind].id for ind in task.depends_on]
            elif isinstance(task, InspectionTask):
                if task.depends_on is None:
                    task.depends_on = [last_drive_to_task.id]
                else:
                    task.depends_on = [self.tasks[ind].id for ind in task.depends_on]

    def _set_unique_id(self) -> None:
        plant_short_name: str = config.get("metadata", "plant_short_name")
        robot_id: str = config.get("metadata", "robot_id")
        now: datetime = datetime.utcnow()
        self.id = f"{plant_short_name.upper()}{robot_id.upper()}{now.strftime('%d%m%Y%H%M%S%f')[:-3]}"

    def __post_init__(self) -> None:
        if self.id is None:
            self._set_unique_id()

        if self.metadata is None:
            self.metadata = MissionMetadata(mission_id=self.id)



import logging
from dataclasses import asdict
from typing import List

from fastapi import Query
from injector import inject
from starlette.responses import JSONResponse

from isar.models.mission import Mission
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.orientation import Orientation
from robot_interface.models.geometry.pose import Pose
from robot_interface.models.geometry.position import Position
from robot_interface.models.mission import DriveToPose


class DriveTo:
    @inject
    def __init__(self, scheduling_utilities: SchedulingUtilities):
        self.logger = logging.getLogger("api")
        self.scheduling_utilities = scheduling_utilities

    def post(
        self,
        x: float = Query(
            ...,
            alias="x-value",
            description="The target x coordinate",
        ),
        y: float = Query(
            ...,
            alias="y-value",
            description="The target y coordinate",
        ),
        z: float = Query(
            ...,
            alias="z-value",
            description="The target z coordinate",
        ),
        q: List[float] = Query(
            [0, 0, 0, 1],
            alias="quaternion",
            description="The target orientation as a quaternion (x,y,z,w)",
        ),
    ):

        ready, response = self.scheduling_utilities.ready_to_start_mission()
        if not ready:
            message, status_code = response
            return JSONResponse(content=asdict(message), status_code=status_code)

        position: Position = Position(x=x, y=y, z=z, frame=Frame.Robot)
        orientation: Orientation = Orientation(
            x=q[0], y=q[1], z=q[2], w=q[3], frame=Frame.Robot
        )
        pose: Pose = Pose(position=position, orientation=orientation, frame=Frame.Robot)

        task: DriveToPose = DriveToPose(pose=pose)
        mission: Mission = Mission([task])

        response = self.scheduling_utilities.start_mission(mission=mission)
        self.logger.info(response)
        message, status_code = response

        return JSONResponse(content=asdict(message), status_code=status_code)



from typing import Optional

from pydantic import BaseModel


class StartResponse(BaseModel):
    message: str
    started: bool
    mission_id: Optional[str] = None


class StopResponse(BaseModel):
    message: str
    stopped: bool



import configparser
import importlib.resources as pkg_resources
from configparser import ConfigParser
from os import getenv

from dotenv import load_dotenv

from isar.config.configuration_error import ConfigurationError


class Config(object):
    def __init__(self):
        load_dotenv()

        self.parser = ConfigParser(interpolation=configparser.ExtendedInterpolation())

        with pkg_resources.path("isar.config", "default.ini") as filepath:
            found_default: bool = self.parser.read(filepath)

        if not found_default:
            raise ConfigurationError(
                f"Failed to import configuration, default: {found_default}"
            )

        robot_package: str = getenv("ROBOT_PACKAGE")
        if robot_package:
            self.parser.set("DEFAULT", "robot_package", robot_package)

    def get(self, section, option):
        return self.parser.get(section, option)

    def getint(self, section, option):
        return self.parser.getint(section, option)

    def getfloat(self, section, option):
        return self.parser.getfloat(section, option)

    def getbool(self, section, option):
        return self.parser.getboolean(section, option)

    def sections(self):
        return self.parser.sections()


config = Config().parser



import logging
from http import HTTPStatus
from logging import Logger
from typing import Union

import click
import uvicorn
from fastapi import FastAPI, Security
from fastapi.middleware.cors import CORSMiddleware
from fastapi.routing import APIRouter
from injector import inject
from pydantic import AnyHttpUrl

from isar.apis.models.models import StartResponse, StopResponse
from isar.apis.schedule.drive_to import DriveTo
from isar.apis.schedule.start_mission import StartMission
from isar.apis.schedule.stop_mission import StopMission
from isar.apis.security.authentication import Authenticator
from isar.config import config


class API:
    @inject
    def __init__(
        self,
        authenticator: Authenticator,
        start_mission: StartMission,
        stop_mission: StopMission,
        drive_to: DriveTo,
        host: str = config.get("DEFAULT", "api_host"),
        port: int = config.getint("DEFAULT", "api_port"),
    ) -> None:

        self.authenticator: Authenticator = authenticator
        self.start_mission: StartMission = start_mission
        self.stop_mission: StopMission = stop_mission
        self.drive_to: DriveTo = drive_to
        self.host: str = host
        self.port: int = port

        self.logger: Logger = logging.getLogger("api")

        self.app: FastAPI = self._create_app()

    def get_app(self) -> FastAPI:
        return self.app

    def run_app(self) -> None:
        uvicorn.run(
            self.app,
            port=self.port,
            host=self.host,
            reload=False,
            log_config=None,
        )

    def _create_app(self) -> FastAPI:
        tags_metadata = [
            {
                "name": "Scheduler",
                "description": "Mission functionality",
            }
        ]
        app = FastAPI(
            openapi_tags=tags_metadata,
            on_startup=[self.authenticator.load_config, self._log_startup_message],
            swagger_ui_oauth2_redirect_url="/oauth2-redirect",
            swagger_ui_init_oauth={
                "usePkceWithAuthorizationCodeGrant": True,
                "clientId": self.authenticator.openapi_client_id,
            },
        )

        if self.authenticator.should_authenticate():
            backend_cors_origins: list[Union[str, AnyHttpUrl]] = [
                f"http://{self.host}:{self.port}"
            ]

            app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in backend_cors_origins],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

        app.include_router(router=self._create_scheduler_router())

        return app

    def _create_scheduler_router(self) -> APIRouter:

        router: APIRouter = APIRouter(tags=["Scheduler"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            "/schedule/start-mission",
            self.start_mission.post,
            methods=["POST"],
            dependencies=[authentication_dependency],
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully started",
                    "model": StartResponse,
                },
                HTTPStatus.NOT_FOUND.value: {
                    "description": "Not found - Mission not found",
                    "model": StartResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Mission already ongoing",
                    "model": StartResponse,
                },
                HTTPStatus.REQUEST_TIMEOUT.value: {
                    "description": "Timeout - Could not contact state machine",
                    "model": StartResponse,
                },
            },
        )

        router.add_api_route(
            "/schedule/stop-mission",
            self.stop_mission.post,
            methods=["POST"],
            dependencies=[authentication_dependency],
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully stopped",
                    "model": StopResponse,
                },
                HTTPStatus.REQUEST_TIMEOUT.value: {
                    "description": "Timeout - Could not contact state machine",
                    "model": StopResponse,
                },
            },
        )
        router.add_api_route(
            "/schedule/drive-to",
            self.drive_to.post,
            methods=["POST"],
            dependencies=[authentication_dependency],
        )

        return router

    def _log_startup_message(self) -> None:
        address_format = "%s://%s:%d/docs"
        message = f"Uvicorn running on {address_format} (Press CTRL+C to quit)"
        protocol = "http"
        color_message = (
            "Uvicorn running on "
            + click.style(address_format, bold=True)
            + " (Press CTRL+C to quit)"
        )

        self.logger.info(
            message,
            protocol,
            self.host,
            self.port,
            extra={"color_message": color_message},
        )



import json
from http import HTTPStatus
from typing import Optional, Tuple

import pytest

from isar.apis.security.authentication import Authenticator
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.models.communication.messages import (
    StartMessage,
    StartMissionMessages,
    StopMissionMessages,
)
from isar.models.communication.messages.stop_message import StopMessage
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from tests.mocks.mission_definition import mock_mission_definition
from tests.mocks.status import mock_status


def mock_check_queue(was_mission_started, state_at_request):
    mock_return_1 = mock_status(
        mission_in_progress=was_mission_started, current_state=state_at_request
    )
    if was_mission_started or state_at_request != "idle":
        mock_return_2 = StartMissionMessages.mission_in_progress()
    else:
        mock_return_2 = StartMissionMessages.success()

    return [mock_return_1, mock_return_2]


def mock_ready_to_start_mission(
    status_code: int,
) -> Tuple[bool, Optional[Tuple[StartMessage, int]]]:
    if status_code == HTTPStatus.OK:
        return True, None
    elif status_code == HTTPStatus.REQUEST_TIMEOUT:
        return (
            False,
            (StartMissionMessages.queue_timeout(), HTTPStatus.REQUEST_TIMEOUT),
        )
    elif status_code == HTTPStatus.CONFLICT:
        return (
            False,
            (StartMissionMessages.mission_in_progress(), HTTPStatus.CONFLICT),
        )
    return True, None


def mock_start_mission(status_code: int) -> Tuple[StartMessage, int]:
    if status_code == HTTPStatus.OK:
        return StartMissionMessages.success(), HTTPStatus.OK
    elif status_code == HTTPStatus.REQUEST_TIMEOUT:
        return (
            StartMissionMessages.queue_timeout(),
            HTTPStatus.REQUEST_TIMEOUT,
        )
    return StartMissionMessages.success(), HTTPStatus.OK


class TestSchedulerRoutes:
    @pytest.mark.parametrize(
        "mission_id, mock_get_mission, expected_exception,"
        "mock_ready_to_start, mock_start, expected_output, expected_status_code",
        [
            (
                12345,
                mock_mission_definition(),
                MissionPlannerError,
                mock_ready_to_start_mission(HTTPStatus.OK),
                mock_start_mission(HTTPStatus.OK),
                StartMissionMessages.mission_not_found(),
                HTTPStatus.NOT_FOUND,
            ),
            (
                1,
                mock_mission_definition(),
                None,
                mock_ready_to_start_mission(HTTPStatus.REQUEST_TIMEOUT),
                mock_start_mission(HTTPStatus.OK),
                StartMissionMessages.queue_timeout(),
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                1,
                mock_mission_definition(),
                None,
                mock_ready_to_start_mission(HTTPStatus.CONFLICT),
                mock_start_mission(HTTPStatus.OK),
                StartMissionMessages.mission_in_progress(),
                HTTPStatus.CONFLICT,
            ),
            (
                1,
                mock_mission_definition(),
                None,
                mock_ready_to_start_mission(HTTPStatus.OK),
                mock_start_mission(HTTPStatus.REQUEST_TIMEOUT),
                StartMissionMessages.queue_timeout(),
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                1,
                mock_mission_definition(),
                None,
                mock_ready_to_start_mission(HTTPStatus.OK),
                mock_start_mission(HTTPStatus.OK),
                StartMissionMessages.success(),
                HTTPStatus.OK,
            ),
        ],
    )
    def test_start_mission(
        self,
        client,
        access_token,
        mocker,
        mission_id,
        mock_get_mission,
        expected_exception,
        mock_ready_to_start,
        mock_start,
        expected_output,
        expected_status_code,
    ):
        mocker.patch.object(
            LocalPlanner,
            "get_mission",
            return_value=mock_get_mission,
            side_effect=expected_exception,
        )
        mocker.patch.object(
            SchedulingUtilities,
            "ready_to_start_mission",
            return_value=mock_ready_to_start,
        )
        mocker.patch.object(
            SchedulingUtilities,
            "start_mission",
            return_value=mock_start,
        )

        response = client.post(
            f"schedule/start-mission?ID={mission_id}",
            headers={"Authorization": "Bearer {}".format(access_token)},
        )

        result_message = json.loads(response.text)
        result = StartMessage(
            message=result_message["message"], started=result_message["started"]
        )
        assert result == expected_output
        assert response.status_code == expected_status_code

    @pytest.mark.parametrize(
        "should_stop, mock_return, expected_output, expected_status_code",
        [
            (
                True,
                QueueTimeoutError(),
                StopMissionMessages.queue_timeout(),
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                True,
                [StopMissionMessages.success()],
                StopMissionMessages.success(),
                HTTPStatus.OK,
            ),
        ],
    )
    def test_stop_mission(
        self,
        client,
        access_token,
        mocker,
        should_stop,
        mock_return,
        expected_output,
        expected_status_code,
    ):
        mocker.patch.object(QueueUtilities, "check_queue", side_effect=mock_return)

        response = client.post(
            "/schedule/stop-mission",
            headers={"Authorization": "Bearer {}".format(access_token)},
        )

        result_message = json.loads(response.text)
        result = StopMessage(
            message=result_message["message"], stopped=result_message["stopped"]
        )

        assert result == expected_output
        assert response.status_code == expected_status_code

    @pytest.mark.parametrize(
        "was_mission_started, state_at_request,expected_status_code,request_params",
        [
            (
                False,
                "idle",
                HTTPStatus.OK,
                {"x": 1, "y": 1, "z": 1, "orientation": "0,0,0,1"},
            ),
            (
                True,
                "idle",
                HTTPStatus.CONFLICT,
                {"x": 1, "y": 1, "z": 1, "orientation": "0,0,0,1"},
            ),
            (
                False,
                "get_mission",
                HTTPStatus.CONFLICT,
                {"x": 1, "y": 1, "z": 1, "orientation": "0,0,0,1"},
            ),
            (
                False,
                "idle",
                HTTPStatus.UNPROCESSABLE_ENTITY,
                {"x": None, "y": 1, "z": 1, "orientation": "0,0,0,1"},
            ),
        ],
    )
    def test_schedule_drive_to(
        self,
        client,
        access_token,
        was_mission_started,
        state_at_request,
        mocker,
        expected_status_code,
        request_params,
    ):
        mocker_return = mock_check_queue(was_mission_started, state_at_request)
        mocker.patch.object(QueueUtilities, "check_queue", side_effect=mocker_return)

        mocker.patch.object(Authenticator, "should_authenticate", return_value=False)

        query_string = (
            f"x-value={request_params['x']}&"
            f"y-value={request_params['y']}&"
            f"z-value={request_params['z']}&"
            f"quaternion={request_params['orientation'][0]}&"
            f"quaternion={request_params['orientation'][2]}&"
            f"quaternion={request_params['orientation'][4]}&"
            f"quaternion={request_params['orientation'][6]}"
        )

        response = client.post(
            f"/schedule/drive-to?{query_string}",
            headers={"Authorization": "Bearer {}".format(access_token)},
        )

        assert response.status_code == expected_status_code



from robot_interface.models.geometry.frame import Frame
from robot_interface.models.geometry.orientation import Orientation
from robot_interface.models.geometry.pose import Pose
from robot_interface.models.geometry.position import Position
from robot_interface.models.mission import DriveToPose, TakeImage


class MockTask:
    @staticmethod
    def drive_to(
        position: Position = Position(x=1, y=1, z=1, frame=Frame.Robot),
        orientation: Orientation = Orientation(
            x=0, y=0, z=0.7071068, w=0.7071068, frame=Frame.Robot
        ),
    ) -> DriveToPose:
        return DriveToPose(
            pose=Pose(position=position, orientation=orientation, frame=Frame.Robot)
        )

    @staticmethod
    def take_image_in_coordinate_direction(
        target: Position = Position(x=1, y=1, z=1, frame=Frame.Robot)
    ) -> TakeImage:
        return TakeImage(target=target)



import logging
import time
from typing import TYPE_CHECKING

from transitions import State

from isar.state_machine.states_enum import States

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Idle(State):
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="idle", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")

    def start(self):
        time.sleep(self.state_machine.sleep_time)
        self.state_machine.update_state()
        self._run()

    def stop(self):
        pass

    def _run(self):
        while True:
            time.sleep(self.state_machine.sleep_time)
            if self.state_machine.should_stop_mission():
                self.state_machine.stop_mission()

            next_state: States = self._get_next_state()
            if not next_state == States.Idle:
                break

        self.state_machine.to_next_state(next_state)

    def _get_next_state(self) -> States:
        (should_start_mission, mission) = self.state_machine.should_start_mission()
        if should_start_mission:
            self.state_machine.start_mission(mission)
            return States.Send

        if self.state_machine.should_send_status():
            self.state_machine.send_status()

        return States.Idle



import json
import logging
import queue
from collections import deque
from copy import deepcopy
from typing import Deque, Optional, Tuple

from injector import Injector, inject
from transitions import Machine
from transitions.core import State

from isar.config import config
from isar.models.communication.messages import (
    StartMissionMessages,
    StopMessage,
    StopMissionMessages,
)
from isar.models.communication.queues.queues import Queues
from isar.models.communication.status import Status
from isar.models.mission import Mission
from isar.services.service_connections.mqtt.mqtt_client import MqttClientInterface
from isar.services.utilities.json_service import EnhancedJSONEncoder
from isar.state_machine.states import Finalize, Idle, Monitor, Off, Send
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions import RobotException
from robot_interface.models.mission.status import TaskStatus
from robot_interface.models.mission.task import Task
from robot_interface.robot_interface import RobotInterface


class StateMachine(object):
    """Handles state transitions for supervisory robot control."""

    @inject
    def __init__(
        self,
        queues: Queues,
        robot: RobotInterface,
        mqtt_client: MqttClientInterface,
        sleep_time: float = config.getfloat("DEFAULT", "fsm_sleep_time"),
        stop_robot_attempts_limit: int = config.getint(
            "DEFAULT", "stop_robot_attempts_limit"
        ),
        transitions_log_length: int = config.getint(
            "DEFAULT", "state_transitions_log_length"
        ),
    ):
        """Initializes the state machine.

        Parameters
        ----------
        queues : Queues
            Queues used for API communication.
        robot : RobotInterface
            Instance of robot interface.
        sleep_time : float
            Time to sleep in between state machine iterations.
        stop_robot_attempts_limit : int
            Maximum attempts to stop the robot when stop command is received
        transitions_log_length : int
            Length of state transition log list.

        """
        self.logger = logging.getLogger("state_machine")

        self.queues = queues
        self.robot = robot
        self.mqtt_client: Optional[MqttClientInterface] = mqtt_client

        self.states = [
            Off(self),
            Idle(self),
            Send(self),
            Monitor(self),
            Finalize(self),
        ]
        self.machine = Machine(
            self,
            states=self.states,
            initial="off",
            queued=True,
        )
        self.stop_robot_attempts_limit = stop_robot_attempts_limit
        self.sleep_time = sleep_time

        self.mission_in_progress: bool = False
        self.current_mission: Mission = Mission(tasks=[])
        self.current_task: Optional[Task] = None
        self.current_task_index: int = -1

        self.current_state: State = States(self.state)  # type: ignore

        self.predefined_mission_id: Optional[int] = None

        self.transitions_log_length = transitions_log_length
        self.transitions_list: Deque[States] = deque([], self.transitions_log_length)

    def begin(self):
        """Starts the state machine.

        Transitions into idle state.

        """
        self._log_state_transition(States.Idle)
        self.to_idle()

    def to_next_state(self, next_state):
        """Transitions state machine to next state."""
        self._log_state_transition(next_state)

        if next_state == States.Idle:
            self.to_idle()
        elif next_state == States.Send:
            self.to_send()
        elif next_state == States.Monitor:
            self.to_monitor()
        elif next_state == States.Finalize:
            self.to_finalize()
        else:
            self.logger.error("Not valid state direction.")

    def update_current_task(self):
        self.current_task_index += 1
        if len(self.current_mission.tasks) > self.current_task_index:
            self.current_task = self.current_mission.tasks[self.current_task_index]
        else:
            self.current_task = None

    def update_state(self):
        """Updates the current state of the state machine."""
        self.current_state = States(self.state)
        self.logger.info(f"State: {self.current_state}")

        payload: str = json.dumps({"state": self.current_state})

        if self.mqtt_client:
            self.mqtt_client.publish(
                topic=config.get("mqtt_topics", "isar_state"),
                payload=payload,
                retain=True,
            )

    def reset_state_machine(self) -> States:
        """Resets the state machine.

        The mission status and progress is reset, and mission schedule
        is emptied.

        Transitions to idle state.

        Returns
        -------
        States
            Idle state.

        """

        self.mission_in_progress = False
        self.current_task = None
        self.current_task_index = -1
        self.current_mission = Mission(tasks=[])

        return States.Idle

    def send_status(self):
        """Communicates state machine status."""
        status = Status(
            mission_in_progress=self.mission_in_progress,
            current_task=self.current_task,
            current_mission=self.current_mission,
            current_state=self.current_state,
        )
        self.queues.mission_status.output.put(deepcopy(status))

    def should_send_status(self) -> bool:
        """Determines if mission status should be sent.

        Returns
        -------
        bool
            True if nonempty queue, false otherwise.

        """
        try:
            send: bool = self.queues.mission_status.input.get(block=False)
            return send
        except queue.Empty:
            return False

    def should_start_mission(self) -> Tuple[bool, Optional[Mission]]:
        """Determines if mission should be started.

        Returns
        -------
        Tuple[bool, Optional[Mission]]
            True if no mission in progress, false otherwise.

        """
        try:
            mission: Mission = self.queues.start_mission.input.get(block=False)
        except queue.Empty:
            return False, None

        if not self.mission_in_progress and mission is not None:
            return True, mission
        elif self.mission_in_progress:
            self.queues.start_mission.output.put(
                deepcopy(StartMissionMessages.mission_in_progress())
            )
            self.logger.info(StartMissionMessages.mission_in_progress())
            return False, None

        return False, None

    def start_mission(self, mission: Mission):
        """Starts a scheduled mission."""
        self.mission_in_progress = True
        self.current_mission = mission
        self.queues.start_mission.output.put(deepcopy(StartMissionMessages.success()))
        self.logger.info(f"Starting new mission: {mission.id}")
        self.log_task_overview(mission=mission)

    def should_stop_mission(self) -> bool:
        """Determines if the running mission should be stopped.

        Returns
        -------
        bool
            True if stop signal is sent and mission is in progress, false otherwise.

        """
        try:
            stop: bool = self.queues.stop_mission.input.get(block=False)
        except queue.Empty:
            return False

        if stop and self.mission_in_progress:
            return True
        elif stop and not self.mission_in_progress:
            message: StopMessage = StopMissionMessages.no_active_missions()
            self.queues.stop_mission.output.put(deepcopy(message))
            self.logger.info(message)
            return False

        return False

    def stop_mission(self):
        """Stops a mission in progress."""
        stop_attempts = 0
        while True:
            try:
                self.robot.stop()
                break
            except RobotException:
                stop_attempts += 1
                if stop_attempts < self.stop_robot_attempts_limit:
                    continue
                self.logger.warning("Failed to stop the robot within maximum attempts!")
                break

        self.mission_in_progress = False
        message: StopMessage = StopMissionMessages.success()
        self.queues.stop_mission.output.put(deepcopy(message))
        self.logger.info(message)

    def publish_task_status(self) -> None:
        """Publishes the current task status to the MQTT Broker"""
        payload: str = json.dumps(
            {
                "task_id": self.current_task.id if self.current_task else None,
                "task_status": self.current_task.status if self.current_task else None,
            },
            cls=EnhancedJSONEncoder,
        )

        self.mqtt_client.publish(
            topic=config.get("mqtt_topics", "isar_task_status"),
            payload=payload,
            retain=True,
        )

    def publish_mission(self) -> None:
        payload: str = json.dumps(
            {"mission": self.current_mission}, cls=EnhancedJSONEncoder
        )

        self.mqtt_client.publish(
            topic=config.get("mqtt_topics", "isar_mission"),
            payload=payload,
            retain=True,
        )

    def _log_state_transition(self, next_state):
        """Logs all state transitions that are not self-transitions."""
        if next_state != self.current_state:
            self.transitions_list.append(next_state)

    def log_task_overview(self, mission: Mission):
        """Log an overview of the tasks in a mission"""
        task_status: str = "\n".join(
            [
                f"{i:>3}  {type(task).__name__:<20} {str(task.id)[:8]:<32} -- {task.status}"
                for i, task in enumerate(mission.tasks)
            ]
        )
        self.logger.info(f"Mission task overview:\n{task_status}")

    def _check_dependencies(self):
        """Check dependencies of previous tasks"""
        if self.current_task and self.current_task.depends_on:
            dependency_tasks = [
                task
                for task in self.current_mission.tasks
                if task.id in self.current_task.depends_on
            ]
            if not all(
                [task.status == TaskStatus.Completed for task in dependency_tasks]
            ):
                return False
        return True


def main(injector: Injector):
    """Starts a state machine instance."""
    state_machine = injector.get(StateMachine)
    state_machine.begin()



from dataclasses import dataclass
from typing import Optional

from isar.models.mission import Mission
from isar.state_machine.states_enum import States
from robot_interface.models.mission import Task, TaskStatus


@dataclass
class Status:
    mission_in_progress: bool
    current_task: Optional[Task]
    current_mission: Mission
    current_state: States



from dataclasses import dataclass, field
from typing import Any, List, Literal, Optional, Union
from uuid import UUID, uuid4

from robot_interface.models.geometry.pose import Pose
from robot_interface.models.geometry.position import Position
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.status import TaskStatus


@dataclass
class Task:
    """
    Base class for all tasks in a mission.
    """

    id: UUID = field(default_factory=uuid4, init=False)
    status: TaskStatus = field(default=TaskStatus.NotStarted, init=False)

    def __str__(self):
        def add_indent(text: str) -> str:
            return "".join("  " + line for line in text.splitlines(True))

        def robot_class_to_pretty_string(obj: Task) -> str:
            log_message: str = ""
            for attr in dir(obj):
                if callable(getattr(obj, attr)) or attr.startswith("__"):
                    continue

                value: Any = getattr(obj, attr)
                try:
                    package_name: Optional[str] = (
                        str(value.__class__).split("'")[1].split(".")[0]
                    )
                except (AttributeError, IndexError) as e:
                    package_name = None

                if package_name == "robot_interface":
                    log_message += (
                        "\n" + attr + ": " + robot_class_to_pretty_string(value)
                    )
                else:
                    log_message += "\n" + str(attr) + ": " + str(value)

            return add_indent(log_message)

        class_name: str = type(self).__name__
        return class_name + robot_class_to_pretty_string(self)


@dataclass
class InspectionTask(Task):
    """
    Base class for all inspection tasks which produce results to be uploaded.
    """

    inspections: List[Inspection] = field(default_factory=list, init=False)
    tag_id: Optional[str] = field(default=None, init=False)

    pass


@dataclass
class MotionTask(Task):
    """
    Base class for all tasks which should cause the robot to move, but not return a result.
    """

    pass


@dataclass
class ContinousInspectionTask(Task):
    """
    Base class for all continous inspection tasks which produce a result to be uploaded.
    """

    pass


@dataclass
class DriveToPose(MotionTask):
    """
    Task which causes the robot to move to the given pose.
    """

    pose: Pose
    name: Literal["drive_to_pose"] = "drive_to_pose"
    depends_on: Optional[List[int]] = None


@dataclass
class DockingProcedure(MotionTask):
    """
    Task which causes the robot to dock or undock
    """

    behavior: Literal["dock, undock"]
    name: Literal["docking_procedure"] = "docking_procedure"


@dataclass
class TakeImage(InspectionTask):
    """
    Task which causes the robot to take an image towards the given coordinate.
    """

    target: Position
    name: Literal["take_image"] = "take_image"
    depends_on: Optional[List[int]] = None


@dataclass
class TakeThermalImage(InspectionTask):
    """
    Task which causes the robot to take a thermal image towards the given coordinate.
    """

    target: Position
    name: Literal["take_thermal_image"] = "take_thermal_image"
    depends_on: Optional[List[int]] = None


TASKS = Union[DriveToPose, DockingProcedure, TakeImage, TakeThermalImage]
